from cefeste.utils import get_categorical_features, get_numerical_features
import warnings


class FeatureAnalysis:
    """Feature Analysis base class."""

    def __init__(self, db, feat_to_check=None):
        """Initialize the class.

        Args:
            db (pd.DataFrame): DataFrame to analyze
            feat_to_check (list, optional): Feature to analyze. If None all are used. Defaults to None.
        """
        self.db = db
        if feat_to_check is None:
            feat_to_check = db.columns

        self.feat_to_check = feat_to_check
        self.categorical_features = get_categorical_features(
            db[feat_to_check],
        )
        self.numerical_features = get_numerical_features(
            db[feat_to_check],
        )

    def get_params(self, params=None):
        """Get a parameter of the class.

        Args:
            params (str, optional): name of the parameter for which the value has to be returned.
                If None all are returned. Defaults to None.

        Raises:
            ValueError: In case the parameters required do not exist.

        Returns:
            Value of the parameter
        """
        if params is None:
            return self.__dict__
        elif params in self.__dict__.keys():
            return self.__dict__[params]
        else:
            raise ValueError(f"Parameters {params} is not a parameter of the class")

    def set_params(self, exceptions=None, **kwargs):
        """Set the parameter of the class.

        Raises:
            ValueError: In case an unknown parameter has been set.
        """
        if exceptions is None:
            exceptions = []
        for k, v in kwargs.items():
            if k in self.__dict__.keys():
                if k in exceptions:
                    warnings.warn(f"Parameter {k} cannot be overwritten")
                else:
                    self.__dict__[k] = v
            else:
                raise ValueError(f"Parameter {k} is not a parameter of the class")

    def eda(self):
        """Generate Exploratory Data Analysis.

        Returns:
            dict: Dictionary with describe or value counts.
        """
        analysis = dict()
        for col in self.numerical_features:
            analysis[col] = self.db[col].describe()
        for col in self.categorical_features:
            analysis[col] = self.db[col].value_counts()
        return analysis

    def __union__(self, attr_name, new_list):
        """It perform the union between the list stored in the attribute named 'attr_name' and 'new_list' and assign the result to the attribute.

        Args:
            attr_name (str): name of the attribute we want to update
            new_list (list): list we want to merge
        """
        old_list = getattr(self, attr_name)
        setattr(self, attr_name, list(set(old_list).union(set(new_list))))

    def __intersection__(self, attr_name, new_list):
        """It perform the inersection between the list stored in the attribute named 'attr_name' and 'new_list' and assign the result to the attribute.

        Args:
            attr_name (str): name of the attribute we want to update
            new_list (list): list we want to intersecate
        """
        old_list = getattr(self, attr_name)
        setattr(self, attr_name, list(set(old_list).intersection(set(new_list))))



import yaml
import pkgutil


def read_config(config_dir="config"):
    """
    Return the project configuration settings.

    Args:
        config_dir (str): Name of directory containing config

    Returns:
        (dict): combined general settings with params settings
    """
    params_config_path = pkgutil.get_data(__name__, f"{config_dir}/params.yaml")
    configurations = yaml.safe_load(params_config_path)

    return configurations



# Utilities
import datetime as dt
from functools import wraps


def get_numerical_features(df, feature_filter=None):
    """It gets a list of numerical features' names from a DataFrame excluding unwanted features' names.

    Args:
        df (pd.DataFrame): input dataframe
        feature_filter (List): list of columns that will be filtered when extracting
            numerical features.

    Returns:
        list: the list of numerical feature names
    """
    if feature_filter is None:
        feature_filter = []

    return [var for var in df.columns if (df[var].dtype != "O" and df[var].dtype != "category") and (var not in feature_filter)]


def get_categorical_features(df, feature_filter=None):
    """It gets a list of categorical features' names from a DataFrame excluding unwanted features' names.

    Args:

    df (pd.DataFrame): input dataframe
    feature_filter (List): list of columns that will be filtered when extracting
        categorical features.

    Returns:
        list: the list of categorical feature names
    """
    if feature_filter is None:
        feature_filter = []

    categorical_feat = [var for var in df.columns if (df[var].dtype == "O" or df[var].dtype == "category") and var not in feature_filter]

    return categorical_feat


def remove_features(df, features_to_remove):
    """Remove a list of features from the dataframe.

    Args:
        df (pd.DataFrame): Dataframe from which the features need to be removed.
        features_to_remove (list): List of strings with the labels of the columns
        to remove.
    Returns:
        (pd.DataFrame): Dataframe without columns to exclude.

    """
    before_removal = df.shape[1]
    features_to_keep = list(set(df.columns) - set(features_to_remove))
    df = df[features_to_keep]
    print(f"Number of features reduced from {before_removal} to {df.shape[1]}")

    return df


def time_step(func):
    """A wrapper function to time a function.

    Args:
        func: a function to be timed.
    Returns:
        wrapper: the time that was used to run the func and the resulting shape will be printed.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        tic = dt.datetime.now()
        result = func(*args, **kwargs)
        time_taken = str(dt.datetime.now() - tic)
        print(f"Completed {func.__name__}, ended at: {dt.datetime.now()}. It took {time_taken}s")
        return result

    return wrapper



# Correlation and VIF
import pandas as pd
import math
import numpy as np
import scipy.stats as stats
from sklearn.linear_model import LinearRegression
from sklearn.metrics import r2_score
from random import seed, randint
from statsmodels.stats.outliers_influence import variance_inflation_factor
from statsmodels.tools.tools import add_constant

from itertools import combinations
from tqdm import tqdm
import sys
from cefeste.utils import get_numerical_features, get_categorical_features


# @time_step
def find_correlated_features(
    df,
    correlation_threshold=0.95,
    feat_to_check=None,
    selection_rule="random",
    random_state=42,
    feat_univ_perf=pd.DataFrame(),
    verbose=False,
    return_selection_history=False,
    return_avg_corr=False,
):
    """Retrieve the list of highly correlated features.

    This function evaluates:
    - Spearman correlation for numerical features;
    - Cramers'V for categorical features;
    - R2 for mixed correlation

    Args:
    df (pd.DataFrame): Dataset under analysis
    correlation_threshold (float): correlation limit to define a couple of features highly correlated
    feat_to_check (list, opt.): list of features to analyse. If empty all are used
    selection_rule(str): 'random' or 'univ_perf'. The rule to define which feature in the couple to keep.
        If 'random' it is randomly chosen and the 'random_state' can be defined.
        If 'univ_perf' than the feature with the highest univariate performance is used.
            In this case 'feat_univ_perf' has to be passed as argument.
    random_state(int, opt.): random seed for 'random' selection_rule
    feat_univ_perf (pd.DataFrame): Dataset containing the features' univariate performance
    verbose (bool): If true print a log of the selection chosen.
    return_selection_history (bool): If True return the Dataset containing the selection choices, otherwise return None.
    return_avg_correlation (bool): If True return an additional Dataset containing the average correlations.

    Returns:
        list: list of features highly correlated.
        pd.Dataframe (or None): Dataset containing the selection choices.
        pd.Dataframe (opt.): Dataset containing the average correlations
    """
    if selection_rule not in ("random", "univ_perf"):
        raise ValueError("selection_rule must be either 'random' or 'univ_perf'")
    if feat_to_check is None:
        feat_to_check = df.columns
    if return_selection_history:
        history_list = [pd.DataFrame()] * 3

    nums = get_numerical_features(df[feat_to_check])
    cats = get_categorical_features(df[feat_to_check])
    len_n = len(nums)
    len_c = len(cats)

    # Aggiungi roba in questo range per printare anche "removing num features", "removing cat feature".. ecc ecc
    pb = tqdm(
        range(int(len_n * (len_n - 1) * 0.5 + len_c * (len_c - 1) * 0.5 + len_n * len_c + 3)),
        file=sys.stdout,
        desc="Calculating correlations between numerical features",
        ncols=100,
        leave=True,
    )
    # For Numerical Feature calculate Spearman and take the absolute values
    if len_n > 1:
        corrs = df[nums].corr(method="spearman").abs().stack().reset_index()
        pb.update(len_n * (len_n - 1) * 0.5)
        pb.set_description("Removing numerical features", refresh=True)
        # Sort correlations
        corrs.columns = ["feat1", "feat2", "value"]
        corrs = corrs.loc[corrs["feat1"] != corrs["feat2"]].sort_values("value", ascending=False)

        # Remove highly correlated couples
        residual_corrs_num, num_history = remove_correlated(
            corrs,
            correlation_threshold=correlation_threshold,
            selection_rule=selection_rule,
            random_state=random_state,
            feat_univ_perf=feat_univ_perf,
            return_selection_history=return_selection_history,
            verbose=verbose,
            prog_bar=pb,
        )
        if return_selection_history:
            history_list[0] = num_history.assign(corr_type="numerical")
        # Finalize the lists to keep/drop
        num_keep_list = residual_corrs_num["feat1"].drop_duplicates().tolist()
        num_drop_list = list(set(nums) - set(num_keep_list))
        pb.update(1)
        if verbose:
            pb.write(f"List of numerical features dropped: {num_drop_list}")
    else:
        residual_corrs_num = pd.DataFrame()
        num_keep_list = nums
        num_drop_list = []
        pb.update(1)
    # For Categorical Features calculate Cramers'V
    if len_c > 1:
        pb.set_description("Calculating correlations between categorical features", refresh=True)
        corrs = pd.DataFrame()
        for var1, var2 in combinations(cats, 2):
            corrs = pd.concat(
                [
                    corrs,
                    pd.DataFrame(
                        {
                            "feat1": [var1],
                            "feat2": [var2],
                            "value": Cramers_pair_correlation(var1, var2, df, False),
                        }
                    ),
                ]
            )
            pb.update(1)
        pb.set_description("Removing categorical features", refresh=True)
        corrs = corrs.sort_values("value", ascending=False)

        # Remove highly correlated couples
        residual_corrs_cat, cat_history = remove_correlated(
            corrs,
            correlation_threshold=correlation_threshold,
            selection_rule=selection_rule,
            random_state=random_state,
            feat_univ_perf=feat_univ_perf,
            return_selection_history=return_selection_history,
            verbose=verbose,
            prog_bar=pb,
        )
        if return_selection_history:
            history_list[1] = cat_history.assign(corr_type="categorical")
        # Finalize the lists to keep/drops
        cat_keep_list = pd.concat([residual_corrs_cat["feat1"], residual_corrs_cat["feat2"]]).drop_duplicates().tolist()
        cat_drop_list = list(set(cats) - set(cat_keep_list))
        pb.update(1)
        if verbose:
            pb.write(f"List of categorical features dropped: {cat_drop_list}")
    else:
        pb.update(1)
        residual_corrs_cat = pd.DataFrame()
        cat_keep_list = cats
        cat_drop_list = []

    # Check mixed correlation
    nums = get_numerical_features(df[num_keep_list])
    cats = get_categorical_features(df[cat_keep_list])
    if (len(nums) > 0) & (len(cats) > 0):
        uu = (len_n * len_c) / (len(nums) * len(cats))
        pb.set_description("Calculating correlations mixed features", refresh=True)
        corrs = pd.DataFrame()
        for var1 in nums:
            for var2 in cats:
                corrs = pd.concat(
                    [
                        corrs,
                        pd.DataFrame(
                            {
                                "feat1": [var1],
                                "feat2": [var2],
                                "value": numerical_categorical_correlation(
                                    df[var2].fillna("Missing"), df[var1].fillna(0)
                                ),
                            }
                        ),
                    ]
                )
                pb.update(uu)
        pb.set_description("Removing mixed features", refresh=True)
        corrs = corrs.sort_values("value", ascending=False)
        # Remove highly correlated couples
        residual_corrs_mix, mix_history = remove_correlated(
            corrs,
            correlation_threshold=correlation_threshold,
            selection_rule=selection_rule,
            random_state=random_state,
            feat_univ_perf=feat_univ_perf,
            return_selection_history=return_selection_history,
            verbose=verbose,
            prog_bar=pb,
        )
        if return_selection_history:
            history_list[2] = mix_history.assign(corr_type="mixed")
        # Finalize the lists to keep/drop
        mixed_keep_list = (
            pd.concat([residual_corrs_mix["feat1"], residual_corrs_mix["feat2"]]).drop_duplicates().tolist()
        )
        mixed_drop_list = list(set(nums + cats) - set(mixed_keep_list))
        pb.set_description("Completed find_correlated_features", refresh=True)
        pb.update(1)
        if verbose:
            pb.write(f"List of mixed features dropped: {mixed_drop_list}")
    else:
        pb.set_description("Completed find_correlated_features", refresh=True)
        pb.update(len_n * len_c + 1)
        residual_corrs_mix = pd.DataFrame()
        mixed_drop_list = []

    if return_selection_history:
        selection_history = pd.concat(history_list).reset_index(drop=True)
    else:
        selection_history = None

    if return_avg_corr:
        corr_avg = (
            pd.concat(
                [
                    residual_corrs_num,
                    residual_corrs_cat,
                    residual_corrs_cat.rename(columns={"feat1": "feat2", "feat2": "feat1"}),
                    residual_corrs_mix,
                    residual_corrs_mix.rename(columns={"feat1": "feat2", "feat2": "feat1"}),
                ]
            )
            .groupby("feat1")["value"]
            .mean()
        )

        return num_drop_list + cat_drop_list + mixed_drop_list, selection_history, corr_avg

    else:
        return num_drop_list + cat_drop_list + mixed_drop_list, selection_history


def Cramers_V(x, y, bias_correction=True):
    """Implementation of the corrected Cramer's V statistic.

    It measure the association between two nominal variables.
    It provides a value between 0 and 1 as output.
    Because Cramer's V can be biased, a bias correction is computed as default.
    The original value can be obtained by setting the bias_correction to False.
    Args:
        x (pd.Series): nominal variable.
        y (pd.Series): nominal variable.
        bias_correction (boolean): flag to compte the corrected Cramer's V.
    Returns:
        float: correlation between the two nominal variables.
    """
    confusion_matrix = pd.crosstab(x, y)
    chi2 = stats.chi2_contingency(confusion_matrix)[0]
    n = confusion_matrix.sum().sum()

    phi2 = chi2 / n

    r, k = confusion_matrix.shape

    CramersV = np.sqrt(phi2 / min((k - 1), (r - 1)))

    if bias_correction:
        phi2corr = max(0, phi2 - ((k - 1) * (r - 1)) / (n - 1))
        rcorr = r - ((r - 1) ** 2) / (n - 1)
        kcorr = k - ((k - 1) ** 2) / (n - 1)
        CramersV = np.sqrt(phi2corr / min((kcorr - 1), (rcorr - 1)))

    return CramersV


def Cramers_pair_correlation(feature_1, feature_2, df, verbose):
    """Call (corrected) Cramers V for a pair of nominal features.

    Args:
        feature_1 (string): label of the first feature.
        feature_2 (string): label of the second feature.
        df (pd.DataFrame): feature matrix.
        verbose (boolean): flag to print more output.

    Returns:
        float: correlation between the tow features
    """
    pair_df = df[[feature_1, feature_2]].dropna()

    try:
        correlation = Cramers_V(pair_df[feature_1], pair_df[feature_2])

    # If the correlation throws an error assume correlation is 0 except if
    # the two features are the same. Then it should be 1.
    except Exception:
        correlation = 0
        if feature_1 == feature_2:
            correlation = 1

    if verbose:
        print(feature_1, feature_2, correlation)

    return correlation


def numerical_categorical_correlation(categ, numerical):
    """Compute the correlation between numerical and categorical features.

    It uses a regression of the OHE categorical variable on the numerical variable.
    Please note this is a proxy.
    Args:
        categ (pd.Series or np.array): Values for the categorical variable.
        numerical (pd.Series or np.array): Values for the numerical variable.

    Returns:
        float with the R-square value.
    """
    dummies = pd.get_dummies(pd.DataFrame(categ))

    linear_regression = LinearRegression().fit(dummies, numerical)

    sqrt_r2 = math.sqrt(max(r2_score(numerical, linear_regression.predict(dummies)), 0))

    return sqrt_r2


def remove_correlated(
    flat_corr_db,
    correlation_threshold=0.95,
    selection_rule="random",
    random_state=42,
    feat_univ_perf=pd.DataFrame(),
    return_selection_history=False,
    selection_history=None,
    verbose=False,
    prog_bar=None,
):
    """Identify and remove from the stacked correlation matrix the highly correlated feature.

    Args:
        flat_corr_db: stacked correlation matrix with columns: 'feat1', 'feat2', 'value'
        correlation_threshold (float): correlation limit to define a couple of features highly correlated
        feat_to_check (list, opt.): list of features to analyse. If empty all are used
        selection_rule(str): 'random' or 'univ_perf'. The rule to define which feature in the couple to keep.
            If 'random' it is randomly chosen and the 'random_state' can be defined.
            If 'univ_perf' than the feature with the highest univariate performance is used.
                'feat_univ_perf' has to be passed as argument.
        random_state(int, opt.): random seed for 'random' selection_rule
        feat_univ_perf (pd.DataFrame): Dataset containing the features' univariate performance
        verbose (bool): If true print a log of the selection chosen.

    Returns:
        pd.DataFrame with the residual correlations
    """
    if (return_selection_history) & (selection_history is None):
        selection_history = pd.DataFrame()
    if flat_corr_db.loc[lambda x: x.value > correlation_threshold].shape[0] > 0:
        if verbose:
            if prog_bar is None:
                print(
                    f"The features correlated are {[flat_corr_db.iloc[0, 0], flat_corr_db.iloc[0, 1]]}, with a correlation of {flat_corr_db.iloc[0, 2]}."
                )
            else:
                prog_bar.write(
                    f"The features correlated are {[flat_corr_db.iloc[0, 0], flat_corr_db.iloc[0, 1]]}, with a correlation of {flat_corr_db.iloc[0, 2]}."
                )

        if selection_rule == "random":
            seed(random_state)
            feat_to_drop = flat_corr_db.iloc[0, randint(0, 1)]
        else:
            perf1 = float(feat_univ_perf.loc[feat_univ_perf["name"] == flat_corr_db.iloc[0, 0]]["perf"])
            perf2 = float(feat_univ_perf.loc[feat_univ_perf["name"] == flat_corr_db.iloc[0, 1]]["perf"])
            if perf1 < perf2:
                feat_to_drop = flat_corr_db.iloc[0, 0]
            else:
                feat_to_drop = flat_corr_db.iloc[0, 1]
        # here i update selection history
        if return_selection_history:
            selection_history = pd.concat(
                [
                    selection_history,
                    pd.DataFrame(
                        {
                            "feat1": flat_corr_db.iloc[0, 0],
                            "feat2": flat_corr_db.iloc[0, 1],
                            "corr": flat_corr_db.iloc[0, 2],
                            "feat_dropped": feat_to_drop,
                        },
                        index=[0],
                    ),
                ]
            )
        if verbose:
            if prog_bar is None:
                print(f"The feature removed is: {feat_to_drop}")
            else:
                prog_bar.write(f"The feature removed is: {feat_to_drop}")
        flat_corr_db = flat_corr_db.loc[
            (flat_corr_db["feat1"] != feat_to_drop) & (flat_corr_db["feat2"] != feat_to_drop)
        ]
        flat_corr_db, selection_history = remove_correlated(
            flat_corr_db,
            selection_rule=selection_rule,
            correlation_threshold=correlation_threshold,
            random_state=random_state,
            feat_univ_perf=feat_univ_perf,
            return_selection_history=return_selection_history,
            selection_history=selection_history,
            verbose=verbose,
            prog_bar=prog_bar,
        )

    return flat_corr_db, selection_history


def find_collinear_feature(df, vif_threshold=5, feat_to_check=None, verbose=True):
    """Find the list of feature with a VIF higher than a threshold in order to avoid multi collinearity.

    Args:
        df (pd.DataFrame): Dataframe on which check the VIF
        vif_threshold (int): vif maximum level
        feat_to_check (list, opt.): list of features to analyse. If empty all are used.
        verbose (bool): If true print a log of the elimination process.

    Returns:
        list: feature that causes multi-collinearity to be dropped

    """
    if feat_to_check is None:
        feat_to_check = df.columns

    nums = get_numerical_features(df[feat_to_check])
    db = df[nums]
    pb = tqdm(range(len(nums)), file=sys.stdout, desc="Removing collinear features", ncols=100, leave=True)
    feat_to_keep = vif_ok_cycle(db, vif_threshold=vif_threshold, verbose=verbose, prog_bar=pb)
    pb.set_description("Completed find_collinear_features", refresh=True)
    pb.update(len(feat_to_keep))
    feat_to_drop = list(set(nums) - set(feat_to_keep))
    return feat_to_drop


def vif_ok_cycle(db, vif_threshold=5, verbose=True, prog_bar=None):
    """Cycle for recursive vif check and feature elimination.

    Args:
        db (pd.DataFrame): Dataframe on which check the VIF
        vif_threshold (int): vif maximum level
        verbose (bool): If true print a log of the elimination process.

    Returns:
        list: feature that ha to be kept
    """
    tmp = add_constant(db)
    results = pd.Series(
        [variance_inflation_factor(tmp.fillna(0).values, i) for i in range(tmp.shape[1])], index=tmp.columns
    )
    results = results[results.index != "const"]

    feat_to_keep = list(results.loc[results < vif_threshold].index)

    if results.max() > vif_threshold:
        feat_to_drop = results.idxmax()
        db = db.drop(columns=feat_to_drop)
        if prog_bar is not None:
            prog_bar.update(1)
        if verbose & (prog_bar is not None):
            prog_bar.write(f"The feature {feat_to_drop} has a VIF of {results.max()}. It has been removed")
        keeps = vif_ok_cycle(db, vif_threshold=vif_threshold, verbose=verbose, prog_bar=prog_bar)
        feat_to_keep = list(set(keeps).union(set(feat_to_keep)))
    return feat_to_keep


# @time_step
def find_collinear_feature_optimized(
    df, vif_threshold=5, feat_to_check=None, verbose=True, optimize=False, optim_Series=None, optim_value_ascending=True
):
    """Find the list of feature with a VIF higher than a threshold in order to avoid multi collinearity.

    Instead of perform the classical analysis, this function is a time-saver if used with optimize parameters.
    Wehn optimize is used, the features are checked in a defined order.

    Args:
        df (pd.DataFrame): Dataframe on which check the VIF
        vif_threshold (int): vif maximum level
        feat_to_check (list, opt.): list of features to analyse. If empty all are used.
        verbose (bool): If true print a log of the elimination process.
        optimize (bool): If true, it requires optim_series and perform VIF analysis in determined order.
        optim_Series (pd.Series): Series that express a value for each feature to be checked
        optim_value_ascending (bool): If True the first feature to be checked is the one with the lowest value.
            Otherwise, the highest one is the first.

    Returns:
        list: feature that causes multi-collinearity to be dropped

    """
    if optim_Series is None:
        optim_Series = pd.Series()

    if optimize:
        if feat_to_check is None:
            feat_to_check = df.columns

        nums = get_numerical_features(df[feat_to_check])
        index_ok = list(set(optim_Series.index) & set(nums))
        optim_Series_reduced = optim_Series[index_ok]
        db = df[nums]

        db_nums = add_constant(db)

        col_order = list(optim_Series_reduced.sort_values(ascending=optim_value_ascending).index)
        feat_to_drop = recursive_vif(db_nums, col_order, vif_threshold=vif_threshold, verbose=verbose)

    else:
        feat_to_drop = find_collinear_feature(
            df, vif_threshold=vif_threshold, feat_to_check=feat_to_check, verbose=verbose
        )

    return feat_to_drop


def recursive_vif(
    df,
    col_order,
    vif_threshold=5,
    verbose=True,
):
    """Perform VIF check recursively, one feature after the other in the predefined order.

    Args:
        df (pd.DataFrame): Dataframe on which check the VIF
        col_order (list): list of feature ordered from the first to be cheked to the latest one.
        vif_threshold (int): vif maximum level
        verbose (bool): If true print a log of the elimination process.


    Returns:
        list: feature that causes multi-collinearity to be dropped
    """
    tmp = df.copy()

    pb = tqdm(col_order, file=sys.stdout, desc="Removing collinear features", ncols=100, leave=True)
    for ix, col in enumerate(pb):
        vif_col = variance_inflation_factor(tmp.fillna(0).values, tmp.columns.get_loc(col))
        if vif_col > vif_threshold:
            tmp = tmp.drop(columns=[col])
            if verbose:
                pb.write(f"The feature {col} has a VIF of {vif_col}. It has been removed")
        if ix == len(col_order) - 1:
            pb.set_description("Completed find_collinear_feature_optimized", refresh=True)

    feat_to_drop = list(set(df.columns) - set(tmp.columns))
    return feat_to_drop



import pandas as pd
import numpy as np
import shap
from sklearn.metrics import roc_auc_score, r2_score, balanced_accuracy_score
from sklearn.model_selection import RandomizedSearchCV, StratifiedKFold, GroupKFold
from sklearn.preprocessing import OneHotEncoder, OrdinalEncoder
from sklearn.compose import ColumnTransformer
import re
from cefeste.utils import remove_features, time_step, get_categorical_features
from functools import reduce
import warnings


@time_step
def Shap_RFE_full(
    X_train,
    y_train,
    model,
    grid,
    cv_funct=RandomizedSearchCV,
    cv_scoring="auto",
    n_iter=20,
    manage_groups=False,
    groups=None,
    cv_type=StratifiedKFold(5, random_state=42, shuffle=True),
    algo_type="classification",
    step_size=0.1,
    min_n_feat_step=5,
    final_n_feature=1,
    verbose=True,
    write_final=False,
    write_substep=False,
    use_ohe=True,
    categorical_features_list_ohe=None,
    dim_cat_threshold=10,
):
    """Recursive Feature Elimination Process with Hyperparameters grid search.

    Args:
        X_train (pd.DataFrame): Feature dataset.
        y_train (pd.Series): target set.
        model: classifier or regressor in sklearn API class.
        grid (dict): hyperparameters grid.
        cv_funct: function or class for the Cross Validation. Defaults to RandomizedSearchCV.
        n_iter (int, optional): number of iteration, i.e. set of hyperparams tested in Cross Validation. Defaults to 20.
        manage_groups (bool, optional): determines if there is a feature whose groups have to be kept joined in CV. Defaults to False.
        groups (pd.Series, optional): feature whose groups have to be kept joined in CV. Defaults to None.
        cv_type : function or class for defining the CV sets. Defaults to StratifiedKFold(5, random_state=42, shuffle=True).
        algo_type (str, optional): "classification", "multiclass", "regression", describes the problem type.
            "classification" has to be used only for binary classification. Defaults to "classification".
        step_size (int or float, optional): determines how many features to remove at each step.
            Fixed int or percentage of total features. Defaults to 0.1.
        min_n_feat_step (int, optional): min number of feature to remove at each step. Defaults to 5.
        final_n_feature (int, optional): number of features of the last model. Defaults to 1.
        verbose (bool, optional): If True print a log of information. Defaults to True.
        write_final (bool, optional): If True it saves in the current directory the final report that can be used for quick start. Defaults to False.
        write_substep (bool, optional): If True it saves in the current directory the report after SHAP_RFE that can be used for quick start. Defaults to False.
        use_ohe (bool, optional): determines whether to use One Hot Encoding on categorical features or not. Defaults to True.
        categorical_features_list_ohe (list, optional): list of categorical features for One Hot Encoding. Defaults to None.
        dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.

    Return:
    pd.DataFrame: Report of the detail and result of each step of FE and Grid Search.
    """
    if groups is None:
        groups = pd.Series()
        
    if algo_type not in ["classification", "regression", "multiclass"]:
        raise ValueError("algo_type argument must be one of ['classification', 'regression', 'multiclass']")  
        
    if (algo_type == "regression") and (str(cv_type.__class__()).startswith("StratifiedKFold")):
            raise ValueError("Fold Cross Validation uncorrect for regression algorithm, use KFold() or GroupKFold()")
            
    if cv_scoring == "auto":
        if algo_type == "classification":
            cv_scoring="roc_auc"
        elif algo_type == "regression":
            cv_scoring="r2"
        elif algo_type == "multiclass":
            cv_scoring="balanced_accuracy"

    # Define the algo
    try:
        CVSel_algo = cv_funct(model, grid, n_iter=n_iter, cv=cv_type, scoring=cv_scoring)
    except Exception:
        CVSel_algo = cv_funct(model, grid, cv=cv_type, scoring=cv_scoring)  # for GridSearchCV

    # Define the groups for Cross Validation
    if manage_groups:
        if len(groups) != len(X_train):
            raise ValueError(
                "dataset to be performed shap and groups-Series don't have the same number of rows ({0},{1})".format(
                    len(X_train), len(groups)
                )
            )
        number_splits = cv_type.n_splits
        CVSel_algo.cv = GroupKFold(number_splits).split(X_train, y_train, groups)
    else:
        number_splits = None

    # Perform Shap Recursive Feature Elimination
    report = Shap_RFE(
        X_train,
        y_train,
        CVSel_algo,
        algo_type=algo_type,
        step_size=step_size,
        min_n_feat_step=min_n_feat_step,
        final_n_feature=final_n_feature,
        verbose=verbose,
        write=write_substep,
        use_ohe=use_ohe,
        categorical_features_list_ohe=categorical_features_list_ohe,
        number_splits=number_splits,
        dim_cat_threshold=dim_cat_threshold,
    )
    # Train the final model
    X_train = X_train[report["feat_select"].iloc[-1]]
    initial_feat_n = X_train.shape[1]
    if manage_groups:
        CVSel_algo.cv = GroupKFold(number_splits).split(X_train, y_train, groups)
    X_train_tsf = X_train.copy()
    if use_ohe:
        cats_feat = list(set(categorical_features_list_ohe) & set(X_train.columns))   
        for col in cats_feat:
            if X_train_tsf[col].nunique() > dim_cat_threshold:
                values_cat = list(
                    X_train_tsf.groupby(col, sort=False)[col].count().sort_values(ascending=False).index
                )[:dim_cat_threshold]
                for val in values_cat:
                    X_train_tsf[col + "_" + str(val)] = (X_train_tsf[col] == val).astype("int")
                X_train_tsf = X_train_tsf.drop(columns=col)
                cats_feat.remove(col)
        preprocessor = ColumnTransformer(
            transformers=[
                ("cat", OneHotEncoder(handle_unknown="ignore", sparse=False), cats_feat),
            ],
            remainder="passthrough",
        )
        preprocessor.fit(X_train_tsf)
        feat_names = preprocessor.get_feature_names_out()
        feat_names = [re.sub(r"((remainder)|(cat))__", "", x) for x in feat_names]
        X_train_tsf = pd.DataFrame(preprocessor.transform(X_train_tsf), columns=feat_names)
    else:
        X_train_tsf[X_train_tsf.select_dtypes(['object']).columns] = X_train_tsf.select_dtypes(['object']).apply(lambda x: x.astype('category'))
    CVSel_algo.fit(X_train_tsf, y_train)

    model = CVSel_algo.best_estimator_
    # Get measure
    if algo_type == "classification":
        train_model_score = roc_auc_score(y_train, CVSel_algo.predict_proba(X_train_tsf)[:, 1])
    elif algo_type == "regression":
        train_model_score = r2_score(y_train, CVSel_algo.predict(X_train_tsf))
    elif algo_type == "multiclass":
        train_model_score = balanced_accuracy_score(y_train, CVSel_algo.predict(X_train_tsf))

    valid_model_score = CVSel_algo.best_score_
    report_step = pd.DataFrame(
        {
            "n_feat": [initial_feat_n],
            "train_score": [train_model_score],
            "valid_score": [valid_model_score],
            "n_feat_to_remove": [0],
            "feat_used": [list(X_train.columns)],
            "feat_to_remove": [[]],
            "feat_select": [list(X_train.columns)],
            "best_estimator": [model],
        }
    )
    report = pd.concat([report, report_step]).reset_index(drop=True)

    if write_final:
        report.to_csv("rsfe_report_final.csv", index=False)

    return report


@time_step
def Shap_RFE(
    X_train,
    y_train,
    CVSel_algo,
    algo_type="classification",
    step_size=0.1,
    min_n_feat_step=5,
    final_n_feature=1,
    verbose=True,
    write=True,
    use_ohe=False,
    categorical_features_list_ohe=None,
    manage_groups=False,
    groups=None,
    number_splits=5,
    dim_cat_threshold=10,
):
    """Recursive Feature Elimination Process with Hyperparameters grid search.

    Args:
        X_train (pd.DataFrame): Feature dataset.
        y_train (pd.Series): target set.
        CVSel_algo: class that include the classifier/regresor, grid search and detail of CV selection.
        algo_type: "classification", "regression" or "multiclass", describes the problem type. Defaults to "classification".
        step_size (int or float): determines how many features to remove at each step.
            Fixed int or percentage of total features. Defaults to 0.1.
        min_n_feat_step (int): min number of feature to remove at each step. Defaults to 5.
        final_n_feature (int): number of features of the last model. Defaults to 1.
        verbose (bool): If True print a log of information. Defaults to True.
        write (bool): If True it saves in the current directory the report that can be used for quick start. Defaults to True.
        use_ohe (bool, optional): determines whether to use One Hot Encoding on categorical features or not. Defaults to False.
        categorical_features_list_ohe (list, optional): list of categorical features for One Hot Encoding. Defaults to None.
        manage_groups (bool, optional): determines if there is a feature whose groups have to be kept joined in CV. Defaults to False.
        groups (pd.Series, optional): feature whose groups have to be kept joined in CV. Defaults to None.
        number_splits (int, optional): number of splits in CV for GroupFold. Defaults to 5.
        dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.


    Return:
    pd.DataFrame: Report of the detail and result of each step of FE and Grid Search.
    """
    initial_feat_n = X_train.shape[1]
    if use_ohe:
        cats_feat = list(set(X_train.columns) & set(categorical_features_list_ohe))

        if algo_type == "classification":
            shap_importance, model, train_model_score, valid_model_score = get_Shap_AUC(
                X_train, y_train, CVSel_algo, categorical_feature=cats_feat, dim_cat_threshold=dim_cat_threshold
            )
        elif algo_type == "regression":
            shap_importance, model, train_model_score, valid_model_score = get_Shap_R2(
                X_train, y_train, CVSel_algo, categorical_feature=cats_feat, dim_cat_threshold=dim_cat_threshold
            )
        elif algo_type == "multiclass":
            shap_importance, model, train_model_score, valid_model_score = get_Shap_BalAcc(
                X_train, y_train, CVSel_algo, categorical_feature=cats_feat, dim_cat_threshold=dim_cat_threshold
            )
    else:
        if algo_type == "classification":
            shap_importance, model, train_model_score, valid_model_score = get_Shap_AUC(
                X_train,
                y_train,
                CVSel_algo,
            )
        elif algo_type == "regression":
            shap_importance, model, train_model_score, valid_model_score = get_Shap_R2(
                X_train,
                y_train,
                CVSel_algo,
            )
        elif algo_type == "multiclass":
            shap_importance, model, train_model_score, valid_model_score = get_Shap_BalAcc(
                X_train,
                y_train,
                CVSel_algo,
            )

    X_train, feat_to_remove = get_new_X(X_train, shap_importance, step_size, min_n_feat_step, final_n_feature)
    report = pd.DataFrame(
        {
            "n_feat": [initial_feat_n],
            "train_score": [train_model_score],
            "valid_score": [valid_model_score],
            "n_feat_to_remove": [len(feat_to_remove)],
            "feat_used": [feat_to_remove + list(X_train.columns)],
            "feat_to_remove": [feat_to_remove],
            "feat_select": [list(X_train.columns)],
            "best_estimator": [model],
        }
    )
    if verbose:
        print(
            f"Train Score: {train_model_score}, Valid Score: {valid_model_score}, features to remove: {feat_to_remove}"
        )
    if X_train.shape[1] > final_n_feature:
        if manage_groups:
            CVSel_algo.cv = GroupKFold(number_splits).split(X_train, y_train, groups)
        report_step = Shap_RFE(
            X_train,
            y_train,
            CVSel_algo,
            algo_type,
            step_size,
            min_n_feat_step,
            final_n_feature,
            verbose,
            write,
            use_ohe,
            categorical_features_list_ohe,
            manage_groups,
            groups,
            number_splits=number_splits,
            dim_cat_threshold=dim_cat_threshold,
        )
        report = pd.concat([report, report_step])

    if write:
        report.to_csv(f"rsfe_report_until{X_train.shape[1]}", index=False)

    return report


def get_Shap(X_train, model, categorical_shap_ohe=False, categorical_feature=None, y_train=None):
    """Get Features' Shapley Value.

    Args:
        X_train (pd.DataFrame): Feature dataset.
        model: the fitted best estimator of the CV Selection.
        categorical_shap_ohe (bool, optional): determines if there are columns coming from OHE. Defaults to False.
        categorical_feature (list, optional): list of categorical features before OHE. Defaults to [].
        y_train (pd.Series, optional): target set. Defaults to None.
    Returns:
        pd.DataFrame: Shap DataFrame.
    """
    if categorical_feature is None:
        categorical_feature = []

    # Get Shap
    feature_names = X_train.columns
    using_features_imporance = False
    try:
        shap_values = shap.TreeExplainer(model).shap_values(X_train)
    except Exception:
        try:
            # For Support Vector Machine and other no-Tree Algorithms
            n_clusters = max(10, min(1000, int(X_train.shape[0] / 100)))
            model.fit(X_train.values, y_train.values)
            shap_values = shap.KernelExplainer(model.predict, shap.kmeans(X_train, n_clusters)).shap_values(X_train)
        except Exception:
            warnings.warn('shap inconsistent with this numpy version. Using feature_importance_ attribute')
            # For computing features importance due to numpy version issues
            feature_importances = model.feature_importances_
            using_features_importance = True
    
    if using_features_importance:
        shap_importance = pd.DataFrame({"feature": feature_names, "shap_importance": feature_importances})
    else:
        # For (multi) classification
        if isinstance(shap_values, list):
            db_list = list()
            for i in range(len(shap_values)):
                db_list.append(
                    pd.DataFrame(
                        {"feature": feature_names, "shap_importance_" + str(i): np.abs(shap_values[i]).mean(axis=0)}
                    )
                )
            shap_importance = (
                reduce(lambda left, right: pd.merge(left, right, how="outer", on="feature"), db_list)
                .set_index("feature")
                .assign(shap_importance=lambda x: x.sum(axis=1))
                .loc[:, "shap_importance"]
                .reset_index()
            )
        # For regression and (some - depending on the classifier) binary classification
        elif isinstance(shap_values, np.ndarray):
            shap_importance = pd.DataFrame({"feature": feature_names, "shap_importance": np.abs(shap_values).mean(axis=0)})

    # For categorical features one-hot-encoded sum the shap values referring the same pivot feature
    if categorical_shap_ohe:
        for i in categorical_feature:
            shap_importance.loc[shap_importance.feature.str.startswith(i)] = shap_importance.loc[
                shap_importance.feature.str.startswith(i)
            ].assign(
                feature=i,
                shap_importance=shap_importance.loc[shap_importance.feature.str.startswith(i), "shap_importance"].sum(),
            )
        shap_importance = shap_importance.drop_duplicates().dropna().reset_index(drop=True)
    shap_importance.sort_values(by=["shap_importance"], ascending=True, inplace=True)
    return shap_importance


def get_Shap_AUC(X_train, y_train, CVSel_algo, categorical_feature=None, dim_cat_threshold=10):
    """Get the AUC of the model and Features' Shapley Value.

    Args:
        X_train (pd.DataFrame): Feature dataset.
        y_train (pd.Series): target set.
        CVSel_algo: class that include the classifier, grid search and detail of CV selection.
        categorical_feature (list, optional): list of categorical feature to be preprocessed. Defaults to None.
        dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.
    Returns:
        (pd.DataFrame, classifier, float, float): Shap DataFrame, Best Classifier, Train and Valid AUC.
    """
    if categorical_feature is None:
        categorical_feature = []

    # Estimate
    categorical_shap_ohe = False
    X_train_tsf = X_train.copy()
    if len(categorical_feature) > 0:
        # Categorical Preprocessing Hot-One-Encoder
        cats_feat = categorical_feature.copy()
        for col in cats_feat:
            if X_train_tsf[col].nunique() > dim_cat_threshold:
                values_cat = list(
                    X_train_tsf.groupby(col, sort=False)[col].count().sort_values(ascending=False).index
                )[:dim_cat_threshold]
                for val in values_cat:
                    X_train_tsf[col + "_" + str(val)] = (X_train_tsf[col] == val).astype("int")
                X_train_tsf = X_train_tsf.drop(columns=col)
                cats_feat.remove(col)
        preprocessor = ColumnTransformer(
            transformers=[
                ("cat", OneHotEncoder(handle_unknown="ignore", sparse=False), cats_feat),
            ],
            remainder="passthrough",
        )
        preprocessor.fit(X_train_tsf)
        feat_names = preprocessor.get_feature_names_out()
        feat_names = [re.sub(r"((remainder)|(cat))__", "", x) for x in feat_names]
        X_train_tsf = pd.DataFrame(preprocessor.transform(X_train_tsf), columns=feat_names)
        categorical_shap_ohe = True

    else:
        X_train_tsf[X_train_tsf.select_dtypes(['object']).columns] = X_train_tsf.select_dtypes(['object']).apply(lambda x: x.astype('category'))
    
    CVSel_algo.fit(X_train_tsf, y_train)

    model = CVSel_algo.best_estimator_

    # Get AUC
    train_model_score = roc_auc_score(y_train, CVSel_algo.predict_proba(X_train_tsf)[:, 1])
    valid_model_score = CVSel_algo.best_score_

    # Get Shap
    shap_importance = get_Shap(X_train_tsf, model, categorical_shap_ohe, categorical_feature, y_train=y_train)

    return shap_importance, model, train_model_score, valid_model_score


def get_Shap_R2(X_train, y_train, CVSel_algo, categorical_feature=None, dim_cat_threshold=10):
    """Get the R squared of the model and Features' Shapley Value.

    Args:
        X_train (pd.DataFrame): Feature dataset.
        y_train (pd.Series): target set.
        CVSel_algo: class that include the regressor, grid search and detail of CV selection.
        categorical_feature (list, optional): list of categorical feature to be preprocessed. Defaults to None.
        dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.
    Returns:
        (pd.DataFrame, classifier, float, float): Shap DataFrame, Best Classifier, Train and Valid R2.

    """
    if categorical_feature is None:
        categorical_feature = []

    # Estimate
    categorical_shap_ohe = False
    X_train_tsf = X_train.copy()
    if len(categorical_feature) > 0:
        # Categorical Preprocessing Hot-One-Encoder
        cats_feat = categorical_feature.copy()
        for col in cats_feat:
            if X_train_tsf[col].nunique() > dim_cat_threshold:
                values_cat = list(
                    X_train_tsf.groupby(col, sort=False)[col].count().sort_values(ascending=False).index
                )[:dim_cat_threshold]
                for val in values_cat:
                    X_train_tsf[col + "_" + str(val)] = (X_train_tsf[col] == val).astype("int")
                X_train_tsf = X_train_tsf.drop(columns=col)
                cats_feat.remove(col)
        preprocessor = ColumnTransformer(
            transformers=[
                ("cat", OneHotEncoder(handle_unknown="ignore", sparse=False), cats_feat),
            ],
            remainder="passthrough",
        )
        preprocessor.fit(X_train_tsf)
        feat_names = preprocessor.get_feature_names_out()
        feat_names = [re.sub(r"((remainder)|(cat))__", "", x) for x in feat_names]
        X_train_tsf = pd.DataFrame(preprocessor.transform(X_train_tsf), columns=feat_names)
        categorical_shap_ohe = True

    else:
        X_train_tsf[X_train_tsf.select_dtypes(['object']).columns] = X_train_tsf.select_dtypes(['object']).apply(lambda x: x.astype('category'))

    CVSel_algo.fit(X_train_tsf, y_train)

    model = CVSel_algo.best_estimator_

    # Get R2
    train_model_score = r2_score(y_train, CVSel_algo.predict(X_train_tsf))
    valid_model_score = CVSel_algo.best_score_

    # Get Shap
    shap_importance = get_Shap(X_train_tsf, model, categorical_shap_ohe, categorical_feature, y_train=y_train)

    return shap_importance, model, train_model_score, valid_model_score


def get_Shap_BalAcc(X_train, y_train, CVSel_algo, categorical_feature=None, dim_cat_threshold=10):
    """Get the Balanced Accuracy of the model and Features' Shapley Value.

    Args:
        X_train (pd.DataFrame): Feature dataset.
        y_train (pd.Series): target set.
        CVSel_algo: class that include the regressor, grid search and detail of CV selection.
        categorical_feature (list, optional): list of categorical feature to be preprocessed. Defaults to [].
        dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.
    Returns:
        (pd.DataFrame, classifier, float, float): Shap DataFrame, Best Classifier, Train and Valid Balanced accuracy.

    """
    if categorical_feature is None:
        categorical_feature = []

    # Estimate
    categorical_shap_ohe = False
    X_train_tsf = X_train.copy()
    if len(categorical_feature) > 0:
        # Categorical Preprocessing Hot-One-Encoder
        cats_feat = categorical_feature.copy()
        for col in cats_feat:
            if X_train_tsf[col].nunique() > dim_cat_threshold:
                values_cat = list(
                    X_train_tsf.groupby(col, sort=False)[col].count().sort_values(ascending=False).index
                )[:dim_cat_threshold]
                for val in values_cat:
                    X_train_tsf[col + "_" + str(val)] = (X_train_tsf[col] == val).astype("int")
                X_train_tsf = X_train_tsf.drop(columns=col)
                cats_feat.remove(col)
        preprocessor = ColumnTransformer(
            transformers=[
                ("cat", OneHotEncoder(handle_unknown="ignore", sparse=False), cats_feat),
            ],
            remainder="passthrough",
        )
        preprocessor.fit(X_train_tsf)
        feat_names = preprocessor.get_feature_names_out()
        feat_names = [re.sub(r"((remainder)|(cat))__", "", x) for x in feat_names]
        X_train_tsf = pd.DataFrame(preprocessor.transform(X_train_tsf), columns=feat_names)
        categorical_shap_ohe = True

    else:
        X_train_tsf[X_train_tsf.select_dtypes(['object']).columns] = X_train_tsf.select_dtypes(['object']).apply(lambda x: x.astype('category'))

    CVSel_algo.fit(X_train_tsf, y_train)

    model = CVSel_algo.best_estimator_
    # Get Balanced Accuracy
    train_model_score = balanced_accuracy_score(y_train, CVSel_algo.predict(X_train_tsf))
    valid_model_score = CVSel_algo.best_score_

    # Get Shap
    shap_importance = get_Shap(X_train_tsf, model, categorical_shap_ohe, categorical_feature, y_train=y_train)

    return shap_importance, model, train_model_score, valid_model_score


def get_new_X(X_train, shap_importance, step_size=0.1, min_n_feat_step=5, final_n_feature=1):
    """Determines the reduced DataFrame excluding the shap-useless features.

    Args:
        X_train (pd.DataFrame): Feature dataset.
        shap_importance(pd.DataFrame): Shap Value DataFrame.
        step_size (int or float, optional): determines how many features to remove at each step.
            Fixed int or percentage of total features. Defaults to 0.1.
        min_n_feat_step (int, optional): min number of feature to remove at each step. Defaults to 5.
        final_n_feature (int, optional): number of features of the last model. Defaults to 1.

    Return:
    (pd.DataFrame, list): Feature Dataset excluding the useless ones, list of removed features.
    """
    # Determine feat to remove
    n_feat = X_train.shape[1]
    if 0 < step_size < 1:
        n_feat_to_remove = np.floor(n_feat * step_size).astype("int")
    elif step_size >= 1 and step_size < step_size:
        n_feat_to_remove = step_size.astype("int")
    else:
        n_feat_to_remove = n_feat - 1

    n_feat_to_remove = min(n_feat - final_n_feature, max(n_feat_to_remove, min_n_feat_step))

    feat_to_remove = list(shap_importance.iloc[0:n_feat_to_remove, 0])
    X_train_red = remove_features(X_train, feat_to_remove)

    return X_train_red, feat_to_remove



# AUC and corr
import pandas as pd
import numpy as np
from tqdm import tqdm
import sys
from sklearn.tree import DecisionTreeClassifier
from sklearn.preprocessing import OneHotEncoder
from sklearn.metrics import roc_auc_score

from cefeste.utils import get_categorical_features
from cefeste.selection.multivariate import numerical_categorical_correlation


# @time_step
def find_not_explanatory(
    X_train, X_test, y_train, y_test, threshold=0.05, algo_type="classification", dim_cat_threshold=10
):
    """Retrieve the list of features that are not useful in the model.

    Args:
        X_train (pd.DataFrame): Feature Dataset for the train set
        X_test (pd.DataFrame): Feature Dataset for the test set
        y_train (pd.Series): target for the train set
        y_test (pd.Series): target for the test set
        threshold(float): minimum performance acceptable. Gini for classification, Correlation for regression.
        algo_type (str): 'classification' or 'regression', 'multiclass'. Define the type of model
        dim_cat_threshold (int): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.
    Returns:
        (list, pd.DataFrame): list of not useful features, dataframe of the performance for each feature
    """
    if algo_type not in ("classification", "regression", "multiclass"):
        raise ValueError("algo_type must be either 'classification' or 'regression' or 'multiclass")
    auc = (1 + threshold) / 2
    if algo_type == "classification":
        return find_not_explanatory_auc(
            X_train, X_test, y_train, y_test, min_auc_tgt=auc, dim_cat_threshold=dim_cat_threshold
        )
    elif algo_type == "regression":
        return find_not_explanatory_corr(X_train, X_test, y_train, y_test, min_corr_tgt=threshold)
    else:
        return find_not_explanatory_auc_multi(
            X_train, X_test, y_train, y_test, min_auc_tgt=auc, dim_cat_threshold=dim_cat_threshold
        )


def find_not_explanatory_auc(X_train, X_test, y_train, y_test, min_auc_tgt=0.525, dim_cat_threshold=10):
    """Retrieve the list of features that are not useful in a classification model.

    Args:
        X_train (pd.DataFrame): Feature Dataset for the train set
        X_test (pd.DataFrame): Feature Dataset for the test set
        y_train (pd.Series): target for the train set
        y_test (pd.Series): target for the test set
        min_auc_tgt (float): minimum AUROC
        dim_cat_threshold (int): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.
    Returns:
        (list, pd.DataFrame): list of not useful features, dataframe of the performance (AUC) for each feature
    """
    with_test = (X_test is not None) & (y_test is not None)
    results = dict()

    train_col_pb = tqdm(X_train.columns, file=sys.stdout, desc="Performing find_not_explanatory", ncols=100, leave=True)
    for ix, col in enumerate(train_col_pb):
        clf = DecisionTreeClassifier(max_depth=5, min_samples_leaf=0.03)
        if col in get_categorical_features(X_train):
            if X_train[col].nunique() > dim_cat_threshold:
                regressor_tr = X_train[[col]].copy()
                if with_test:
                    regressor_te = X_test[[col]].copy()
                values_cat = list(
                    regressor_tr.groupby(col, sort=False)[col].count().sort_values(ascending=False).index
                )[:dim_cat_threshold]
                for val in values_cat:
                    regressor_tr[col + "_" + str(val)] = (regressor_tr[col] == val).astype("int")
                    if with_test:
                        regressor_te[col + "_" + str(val)] = (regressor_te[col] == val).astype("int")
                regressor_tr = regressor_tr.drop(columns=col)
                if with_test:
                    regressor_te = regressor_te.drop(columns=col)
            else:
                ohe = OneHotEncoder(sparse=False, handle_unknown="ignore")
                regressor_tr = ohe.fit_transform(X_train[[col]])
                if with_test:
                    regressor_te = ohe.transform(X_test[[col]])
        else:
            regressor_tr = X_train[[col]].fillna(X_train[col].median())
            if with_test:
                regressor_te = X_test[[col]].fillna(X_test[col].median())
        if ix == len(X_train.columns) - 1:
            train_col_pb.set_description("Completed find_not_explanatory", refresh=True)

        clf.fit(regressor_tr, y_train)
        y_pred_tr = clf.predict_proba(regressor_tr)[:, 1]
        if with_test:
            y_pred_te = clf.predict_proba(regressor_te)[:, 1]
            results[col] = [roc_auc_score(y_train, y_pred_tr), roc_auc_score(y_test, y_pred_te)]
        else:
            results[col] = roc_auc_score(y_train, y_pred_tr)

    drop_list = [x for x, y in results.items() if np.min(y) < min_auc_tgt]
    perf = pd.DataFrame()
    perf["name"] = results.keys()
    perf["perf"] = results.values()
    if with_test:
        perf[["train", "test"]] = perf.perf.apply(pd.Series)
        perf["perf"] = perf[["train", "test"]].min(axis=1)

    return drop_list, perf


def find_not_explanatory_auc_multi(X_train, X_test, y_train, y_test, min_auc_tgt=0.525, dim_cat_threshold=10):
    """Retrieve the list of features that are not useful in a classification model.

    Args:
        X_train (pd.DataFrame): Feature Dataset for the train set
        X_test (pd.DataFrame): Feature Dataset for the test set
        y_train (pd.Series): target for the train set
        y_test (pd.Series): target for the test set
        min_auc_tgt (float): minimum AUROC
        dim_cat_threshold (int): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
            Defaults to 10.
    Returns:
        (list, pd.DataFrame): list of not useful features, dataframe of the performance (AUC) for each feature
    """
    with_test = (X_test is not None) & (y_test is not None)
    results = dict()

    train_col_pb = tqdm(X_train.columns, file=sys.stdout, desc="Performing find_not_explanatory", ncols=100, leave=True)
    for ix, col in enumerate(train_col_pb):
        clf = DecisionTreeClassifier(max_depth=5, min_samples_leaf=0.03)
        if col in get_categorical_features(X_train):
            if X_train[col].nunique() > dim_cat_threshold:
                regressor_tr = X_train[[col]].copy()
                if with_test:
                    regressor_te = X_test[[col]].copy()
                values_cat = list(
                    regressor_tr.groupby(col, sort=False)[col].count().sort_values(ascending=False).index
                )[:dim_cat_threshold]
                for val in values_cat:
                    regressor_tr[col + "_" + str(val)] = (regressor_tr[col] == val).astype("int")
                    if with_test:
                        regressor_te[col + "_" + str(val)] = (regressor_te[col] == val).astype("int")
                regressor_tr = regressor_tr.drop(columns=col)
                if with_test:
                    regressor_te = regressor_te.drop(columns=col)
            else:
                ohe = OneHotEncoder(sparse=False, handle_unknown="ignore")
                regressor_tr = ohe.fit_transform(X_train[[col]])
                if with_test:
                    regressor_te = ohe.transform(X_test[[col]])
        else:
            regressor_tr = X_train[[col]].fillna(X_train[col].median())
            if with_test:
                regressor_te = X_test[[col]].fillna(X_test[col].median())

        clf.fit(regressor_tr, y_train)
        y_pred_tr = clf.predict_proba(regressor_tr)
        if with_test:
            y_pred_te = clf.predict_proba(regressor_te)
            results[col] = [
                roc_auc_score(y_train, y_pred_tr, multi_class="ovr"),
                roc_auc_score(y_test, y_pred_te, multi_class="ovr"),
            ]
        else:
            results[col] = roc_auc_score(y_train, y_pred_tr, multi_class="ovr")
        if ix == len(X_train.columns) - 1:
            train_col_pb.set_description("Completed find_not_explanatory", refresh=True)

    drop_list = [x for x, y in results.items() if np.min(y) < min_auc_tgt]
    perf = pd.DataFrame()
    perf["name"] = results.keys()
    perf["perf"] = results.values()
    if with_test:
        perf[["train", "test"]] = perf.perf.apply(pd.Series)
        perf["perf"] = perf[["train", "test"]].min(axis=1)

    return drop_list, perf


def find_not_explanatory_corr(X_train, X_test, y_train, y_test, min_corr_tgt=0.05):
    """Retrieve the list of features that are not useful in a regression model.

    Args:
        X_train (pd.DataFrame): Feature Dataset for the train set
        X_test (pd.DataFrame): Feature Dataset for the test set
        y_train (pd.Series): target for the train set
        y_test (pd.Series): target for the test set
        min_corr_tgt (float): minimum correlation
    Returns:
        (list, pd.DataFrame): list of not useful features, dataframe of the performance (Correlation) for each feature
    """
    # to be done
    with_test = (X_test is not None) & (y_test is not None)
    results = dict()

    train_col_pb = tqdm(X_train.columns, file=sys.stdout, desc="Performing find_not_explanatory", ncols=100, leave=True)
    for ix, col in enumerate(train_col_pb):
        if col in get_categorical_features(X_train):
            if with_test:
                results[col] = [
                    numerical_categorical_correlation(X_train[col].fillna(0), y_train),
                    numerical_categorical_correlation(X_test[col].fillna(0), y_test),
                ]
            else:
                results[col] = numerical_categorical_correlation(X_train[col].fillna(0), y_train)
        else:
            if with_test:
                results[col] = [
                    np.abs(X_train[col].fillna(0).corr(y_train)),
                    np.abs(X_test[col].fillna(0).corr(y_test)),
                ]
            else:
                results[col] = np.abs(X_train[col].fillna(0).corr(y_train))
        if ix == len(X_train.columns) - 1:
            train_col_pb.set_description("Completed find_not_explanatory", refresh=True)

    drop_list = [x for x, y in results.items() if np.min(y) < min_corr_tgt]
    perf = pd.DataFrame()
    perf["name"] = results.keys()
    perf["perf"] = results.values()
    if with_test:
        perf[["train", "test"]] = perf.perf.apply(pd.Series)
        perf["perf"] = perf[["train", "test"]].min(axis=1)

    return drop_list, perf



from cefeste import FeatureAnalysis
from cefeste.utils import remove_features, get_numerical_features
from cefeste.selection.explanatory import find_not_explanatory
from cefeste.selection.multivariate import find_collinear_feature_optimized, find_correlated_features
from cefeste.selection.univariate import (
    find_constant_features,
    find_high_topcat_features,
    find_low_nvalues_features,
    find_missing_features,
    find_unstable_psi_features,
)

import warnings
from itertools import combinations
from functools import reduce

import pandas as pd
from numpy import nan


class FeatureSelection(FeatureAnalysis):
    """Child class of FeatureAnalysis for Feature Selection."""

    def __init__(
        self,
        # DB / Feature Parameters
        db,
        feat_to_check=None,
        target_col=None,
        algo_type="auto",
        sample_col=None,
        sample_train_value=None,
        # Univariate Analysis Parameters
        min_unique_val=3,
        max_pct_missing=0.9,
        max_pct_mfv=0.95,
        max_psi=0.2,
        psi_nbins=20,
        psi_bin_min_pct=0.02,
        # Univariate Explanatory Power Parameters
        explanatory_threshold=0.05,
        # Multivariate Analysis Parameters
        correlation_threshold=0.95,
        selection_rule="random",
        vif_threshold=5,
        collinear_optimize=False,
        return_selection_history=True,
        dim_cat_threshold=10,
        # Generic Parameters
        random_state=42,
        verbose=True,
    ):
        """Feature Selection Class.

        Args:
            db (pd.DataFrame): DataFrame to analyze
            feat_to_check (list, optional): Feature to analyze. If None all are used. Defaults to None.
            target_col (str, optional): Name of the target column. Defaults to None.
            algo_type (str, optional): Type of algo. It should be one of the following:
                'auto', 'regression', 'multiclass', 'classification', 'unsupervised'. Defaults to "auto".
            sample_col (str, optional): Name of the columns that indicates the samples (train, test). Defaults to None.
            sample_train_value (str, optional): Value of the sample column that indicates the train set. Defaults to None.
            min_unique_val (int, optional): Minimum number of unique value to be used by find_low_values. Defaults to 3.
            max_pct_missing (float, optional): Max percentage of missing to be used by find_missing filter. Defaults to 0.9.
            max_pct_mfv (float, optional): Max percentage of the most frequent value. Defaults to 0.95.
            max_psi (float, optional): Max PSI. Defaults to 0.2.
            psi_nbins (int, optional): Number of bins to caluclate PSI. Defaults to 20.
            psi_bin_min_pct (float, optional): Min percentage of observation per Bin when calculating psi. Defaults to 0.02.
            explanatory_threshold (float, optional): Threshold for explanatory power. Defaults to 0.05.
            correlation_threshold (float, optional): Correlation threshold. Defaults to 0.95.
            selection_rule (str, optional): Selection rule for correlation. Defaults to "random".
            vif_threshold (int, optional): VIF Threshold. Defaults to 5.
            collinear_optimize (bool, optional): Flag for using the optimized version of collinear check. Defaults to False.
            dim_cat_threshold (int, optional): cardinality threshold for categorical variables to apply or not 'simplified' OHE.
                Defaults to 10.
            random_state (int, optional): random state. Defaults to 42.
            verbose (bool, optional): Boolean to print more information. Defaults to True.

        """
        super().__init__(db, feat_to_check)
        self.feat_to_check = list(set(self.feat_to_check) - set([target_col, sample_col]))
        # Target check
        if (target_col in db.columns) | (target_col is None):
            self.target_col = target_col
        else:
            raise ValueError(f"{target_col} not in DataFrame")

        # Check the target var and set the algo type
        if algo_type not in ["auto", "regression", "classification", "multiclass", "unsupervised"]:
            raise ValueError(
                f"{algo_type} is not a valid algo_type. It should be one of the following:\n ['auto', 'regression', 'classification', 'multiclass', 'unsupervised']"
            )

        if target_col is not None:
            vals = db[target_col].nunique()
            if vals == 1:
                raise ValueError(f"The target column {target_col} selected is constant")
            elif algo_type != "auto":
                self.algo_type = algo_type
            elif vals == 2:
                self.algo_type = "classification"
            elif vals < 11:
                self.algo_type = "multiclass"
            else:
                self.algo_type = "regression"
        else:
            self.algo_type = "unsupervised"

        # Sample column check
        if (sample_col in db.columns) | (sample_col is None):
            self.sample_col = sample_col
        else:
            raise ValueError(f"{sample_col} not in DataFrame")
        # Save the list of samples in db
        if sample_col is not None:
            self._sample_list = db[sample_col].unique()
            self._n_sample = db[sample_col].nunique()
        else:
            self._sample_list = []
            self._n_sample = 0
        # Sample Train value
        if (sample_train_value is not None) & (sample_train_value not in self._sample_list):
            raise ValueError(
                f"The value {sample_train_value} set for parameter sample_train_value is not in {sample_col}."
            )
        else:
            self.sample_train_value = sample_train_value

        # Parameters
        self.min_unique_val = min_unique_val
        self.max_pct_missing = max_pct_missing
        self.max_pct_mfv = max_pct_mfv
        self.max_psi = max_psi
        self.psi_nbins = psi_nbins
        self.psi_bin_min_pct = psi_bin_min_pct
        self.explanatory_threshold = explanatory_threshold
        self.correlation_threshold = correlation_threshold
        self.selection_rule = selection_rule
        self.vif_threshold = vif_threshold
        self.collinear_optimize = collinear_optimize
        self.random_state = random_state
        self.verbose = verbose
        self.return_selection_history = return_selection_history
        self.dim_cat_threshold = dim_cat_threshold

        # Initialize the attributes as empty lists/dataframes
        self._constant_features = []
        self._missing_features = []
        self._highly_concentrated_features = []
        self._low_values_features = []
        self._unstable_features = []
        self._unexplanatory_features = []
        self._correlated_features = []
        self._collinear_features = []
        self._selected_features = self.feat_to_check
        self._filtered_out_features = []
        self._funnel_df = pd.DataFrame(
            {
                "Step_Description": "Initial feat to check",
                "Col_Removed": 0,
                "Col_Kept": len(self.feat_to_check),
                "Params": nan,
            },
            index=[0],
        ).rename_axis("Step_Number")

        # Initialize the attributes as False
        self.filter_constant = False
        self.filter_missing = False
        self.filter_missing = False
        self.filter_highly_concentrated = False
        self.filter_low_values = False
        self.filter_unstable = False
        self.filter_unexplanatory = False
        self.filter_correlated = False
        self.filter_collinear = False

        self._perf_db = pd.DataFrame()
        self._selection_history = pd.DataFrame()

    def run(
        self,
        filter_constant=True,
        filter_missing=True,
        filter_highly_concentrated=True,
        filter_low_values=False,
        filter_unstable=True,
        filter_unexplanatory=True,
        filter_correlated=True,
        filter_collinear=True,
        **kwargs,
    ):
        """Perform the filtering.

        Args:
            filter_constant (bool, optional): _description_. Defaults to True.
            filter_missing (bool, optional): _description_. Defaults to True.
            filter_highly_concentrated (bool, optional): _description_. Defaults to True.
            filter_low_values (bool, optional): _description_. Defaults to False.
            filter_unstable (bool, optional): _description_. Defaults to True.
            filter_unexplanatory (bool, optional): _description_. Defaults to True.
            filter_correlated (bool, optional): _description_. Defaults to True.
            filter_collinear (bool, optional): _description_. Defaults to True.

        Returns:
            pd.DataFrame: filtered db
        """
        self.set_params(exceptions="feat_to_check", **kwargs)
        # for k, v in kwargs.items():
        #    self.__dict__[k] = v

        db = self.db
        db.reset_index(inplace=True, drop=True)

        if self.sample_col is not None:
            df, sample_series = self.db.drop(columns=self.sample_col), self.db[self.sample_col]
        else:
            df = db

        if self.target_col is not None:
            X, y = df.drop(columns=self.target_col), df[self.target_col]
        else:
            X = df
        X = X[self._selected_features]
        # self._X_original = X

        if self.sample_train_value is not None:
            X_train = X.loc[sample_series == self.sample_train_value]
        else:
            X_train = X

        if filter_constant:
            _constant_features = find_constant_features(
                X_train,
            )
            self.__union__("_constant_features", _constant_features)
            self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                {
                    "Step_Description": "Constant",
                    "Col_Removed": len(_constant_features),
                    "Col_Kept": len(X_train.columns) - len(_constant_features),
                    "Params": nan,
                }
            )
            X_train = remove_features(X_train, _constant_features)

        if filter_missing:
            _missing_features = find_missing_features(
                X_train,
            )
            self.__union__("_missing_features", _missing_features)
            self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                {
                    "Step_Description": "Missing",
                    "Col_Removed": len(_missing_features),
                    "Col_Kept": len(X_train.columns) - len(_missing_features),
                    "Params": {"max_pct_missing": self.max_pct_missing},
                }
            )
            X_train = remove_features(X_train, _missing_features)

        if filter_highly_concentrated:
            _highly_concentrated_features = find_high_topcat_features(X_train, max_pct_mfv=self.max_pct_mfv)
            self.__union__("_highly_concentrated_features", _highly_concentrated_features)
            self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                {
                    "Step_Description": "Highly Concentrated",
                    "Col_Removed": len(_highly_concentrated_features),
                    "Col_Kept": len(X_train.columns) - len(_highly_concentrated_features),
                    "Params": {"max_pct_mfv": self.max_pct_mfv},
                }
            )
            X_train = remove_features(X_train, _highly_concentrated_features)

        if filter_low_values:
            _low_values_features = find_low_nvalues_features(X_train, min_unique_val=self.min_unique_val)
            self.__union__("_low_values_features", _low_values_features)
            self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                {
                    "Step_Description": "Low Values",
                    "Col_Removed": len(_low_values_features),
                    "Col_Kept": len(X_train.columns) - len(_low_values_features),
                    "Params": {"min_unique_val": self.min_unique_val},
                }
            )
            X_train = remove_features(X_train, _low_values_features)

        X = X[X_train.columns]

        if filter_unstable:
            if self.sample_col is None:
                filter_unstable = False
                warnings.warn("filter unstable not performed since no sample columns defined")
            elif self._n_sample <= 1:
                filter_unstable = False
                warnings.warn("filter unstable not performed since sample columns defined is constant")
            else:
                sample_comb = list(combinations(self._sample_list, 2))
                unstable_dict = {}
                for comb in sample_comb:
                    if (self.sample_train_value in comb) | (self.sample_train_value is None):
                        base = X.loc[sample_series == comb[0]]
                        compare = X.loc[sample_series == comb[1]]
                        unstable_dict[comb] = find_unstable_psi_features(
                            # evetually add a parameter to specify wich combination we are currenty considering
                            base,
                            compare,
                            max_psi=self.max_psi,
                            psi_bin_min_pct=self.psi_bin_min_pct,
                            psi_nbins=self.psi_nbins,
                        )

                _unstable_features = list(reduce(lambda x, y: set(x) | set(y), unstable_dict.values()))
                self.__union__("_unstable_features", _unstable_features)
                self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                    {
                        "Step_Description": "Unstable",
                        "Col_Removed": len(_unstable_features),
                        "Col_Kept": len(X_train.columns) - len(_unstable_features),
                        "Params": {
                            "max_psi": self.max_psi,
                            "psi_bin_min_pct": self.psi_bin_min_pct,
                            "psi_nbins": self.psi_nbins,
                        },
                    }
                )
                X = remove_features(X, _unstable_features)

        if filter_unexplanatory:
            if self.target_col is None:
                filter_unexplanatory = False
                warnings.warn("filter unexplanatory not performed since no target columns defined")
            elif (self.sample_col is None) | (self._n_sample <= 1):
                warnings.warn("filter unexplanatory considering the whole dataset belonging to the same split")
                _unexplanatory_features, _perf_db = find_not_explanatory(
                    X,
                    None,
                    y,
                    None,
                    threshold=self.explanatory_threshold,
                    algo_type=self.algo_type,
                    dim_cat_threshold=self.dim_cat_threshold,
                )
                self._perf_db = _perf_db
                self.__union__("_unexplanatory_features", _unexplanatory_features)
                self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                    {
                        "Step_Description": "Unexplanatory",
                        "Col_Removed": len(_unexplanatory_features),
                        "Col_Kept": len(X.columns) - len(_unexplanatory_features),
                        "Params": {
                            "threshold": self.explanatory_threshold,
                            "algo_type": self.algo_type,
                            "dim_cat_threshold": self.dim_cat_threshold,
                        },
                    }
                )
                X = remove_features(X, _unexplanatory_features)
            else:
                sample_comb = list(combinations(self._sample_list, 2))
                unexpl_dict = {}
                perf_dict = {}
                for comb in sample_comb:
                    if (self.sample_train_value in comb) | (self.sample_train_value is None):
                        baseX = X.loc[sample_series == comb[0]]
                        compareX = X.loc[sample_series == comb[1]]
                        basey = y.loc[sample_series == comb[0]]
                        comparey = y.loc[sample_series == comb[1]]
                        unexpl_dict[comb], _perf_ = find_not_explanatory(
                            baseX,
                            compareX,
                            basey,
                            comparey,
                            threshold=self.explanatory_threshold,
                            algo_type=self.algo_type,
                            dim_cat_threshold=self.dim_cat_threshold,
                        )
                        perf_dict[comb] = _perf_[["name", "perf"]].copy()

                _unexplanatory_features = list(reduce(lambda x, y: set(x) | set(y), unexpl_dict.values()))
                perf_db = reduce(
                    lambda x, y: x.merge(y, left_index=True, right_index=True, how="outer"), perf_dict.values()
                )
                perf_db.set_index("name", inplace=True)
                perf_db["perf"] = perf_db.min(axis=1)
                _perf_db = perf_db[["perf"]].reset_index()
                self._perf_db = _perf_db
                self.__union__("_unexplanatory_features", _unexplanatory_features)
                self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                    {
                        "Step_Description": "Unexplanatory",
                        "Col_Removed": len(_unexplanatory_features),
                        "Col_Kept": len(X.columns) - len(_unexplanatory_features),
                        "Params": {
                            "threshold": self.explanatory_threshold,
                            "algo_type": self.algo_type,
                            "dim_cat_threshold": self.dim_cat_threshold,
                        },
                    }
                )
                X = remove_features(X, _unexplanatory_features)

        X_train = X_train[X.columns]

        if (filter_correlated) & (X_train.shape[1] > 1):
            _correlated_features, selection_history, self._avg_corr = find_correlated_features(
                X_train,
                correlation_threshold=self.correlation_threshold,
                selection_rule=self.selection_rule,
                random_state=self.random_state,
                feat_univ_perf=self._perf_db,
                return_selection_history=self.return_selection_history,
                verbose=self.verbose,
                return_avg_corr=True,
            )
            if self.return_selection_history:
                self._selection_history = pd.concat([self._selection_history, selection_history])
            self.__union__("_correlated_features", _correlated_features)
            self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                {
                    "Step_Description": "Correlated",
                    "Col_Removed": len(_correlated_features),
                    "Col_Kept": len(X_train.columns) - len(_correlated_features),
                    "Params": {
                        "correlation_threshold": self.correlation_threshold,
                        "selection_rule": self.selection_rule,
                        "random_state": self.random_state,
                    },
                }
            )
            X_train = remove_features(X_train, _correlated_features)

        if (filter_collinear) & (len(get_numerical_features(X_train)) > 1):
            if (not self.collinear_optimize) | (not (filter_correlated or filter_unexplanatory)):
                self.collinear_optimize = False
                optim_Series = None
                optim_value_ascending = True

            elif filter_correlated:
                optim_Series = self._avg_corr
                optim_value_ascending = False
            else:
                optim_Series = self._perf_db.set_index("name")["perf"]
                optim_value_ascending = True

            _collinear_features = find_collinear_feature_optimized(
                X_train,
                vif_threshold=self.vif_threshold,
                verbose=self.verbose,
                optimize=self.collinear_optimize,
                optim_Series=optim_Series,
                optim_value_ascending=optim_value_ascending,
            )
            self.__union__("_collinear_features", _collinear_features)
            self._funnel_df.loc[self._funnel_df.shape[0]] = pd.Series(
                {
                    "Step_Description": "Collinear",
                    "Col_Removed": len(_collinear_features),
                    "Col_Kept": len(X_train.columns) - len(_collinear_features),
                    "Params": {
                        "vif_threshold": self.vif_threshold,
                        "optimize": self.collinear_optimize,
                        # I did not put optim_value_ascending cause it is decided automatically
                    },
                }
            )
            X_train = remove_features(X_train, _collinear_features)

        self.filter_constant |= filter_constant
        self.filter_missing |= filter_missing
        self.filter_highly_concentrated |= filter_highly_concentrated
        self.filter_low_values |= filter_low_values
        self.filter_unstable |= filter_unstable
        self.filter_unexplanatory |= filter_unexplanatory
        self.filter_correlated |= filter_correlated
        self.filter_collinear |= filter_collinear

        self.__intersection__("_selected_features", list(X_train.columns))
        self.__union__("_filtered_out_features", list(set(self.feat_to_check) - set(X_train.columns)))
        return

    def find_constant_features(self, **kwargs):
        """Perform the filtering of constant features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=True,
            filter_missing=False,
            filter_highly_concentrated=False,
            filter_low_values=False,
            filter_unstable=False,
            filter_unexplanatory=False,
            filter_correlated=False,
            filter_collinear=False,
            **kwargs,
        )
        return

    def find_missing_features(self, **kwargs):
        """Perform the filtering of missing features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=False,
            filter_missing=True,
            filter_highly_concentrated=False,
            filter_low_values=False,
            filter_unstable=False,
            filter_unexplanatory=False,
            filter_correlated=False,
            filter_collinear=False,
            **kwargs,
        )
        return

    def find_high_topcat_fetures(self, **kwargs):
        """Perform the filtering of high_topcat features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=False,
            filter_missing=False,
            filter_highly_concentrated=True,
            filter_low_values=False,
            filter_unstable=False,
            filter_unexplanatory=False,
            filter_correlated=False,
            filter_collinear=False,
            **kwargs,
        )
        return

    def find_low_nvalues_features(self, **kwargs):
        """Perform the filtering of low_nvalues features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=False,
            filter_missing=False,
            filter_highly_concentrated=False,
            filter_low_values=True,
            filter_unstable=False,
            filter_unexplanatory=False,
            filter_correlated=False,
            filter_collinear=False,
        )
        return

    def find_unstable_psi_features(self, **kwargs):
        """Perform the filtering of unstable features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=False,
            filter_missing=False,
            filter_highly_concentrated=False,
            filter_low_values=False,
            filter_unstable=True,
            filter_unexplanatory=False,
            filter_correlated=False,
            filter_collinear=False,
            **kwargs,
        )
        return

    def find_not_explanatory(self, **kwargs):
        """Perform the filtering of unexplanatory features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=False,
            filter_missing=False,
            filter_highly_concentrated=False,
            filter_low_values=False,
            filter_unstable=False,
            filter_unexplanatory=True,
            filter_correlated=False,
            filter_collinear=False,
            **kwargs,
        )
        return

    def find_correlated_features(self, **kwargs):
        """Perform the filtering of correlated features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=False,
            filter_missing=False,
            filter_highly_concentrated=False,
            filter_low_values=False,
            filter_unstable=False,
            filter_unexplanatory=False,
            filter_correlated=True,
            filter_collinear=False,
            **kwargs,
        )
        return

    def find_collinear_feature_optimized(self, **kwargs):
        """Perform the filtering of collinear features.

        It simply call run() method with the correspondent argument as True.
        """
        self.run(
            filter_constant=False,
            filter_missing=False,
            filter_highly_concentrated=False,
            filter_low_values=False,
            filter_unstable=False,
            filter_unexplanatory=False,
            filter_correlated=False,
            filter_collinear=True,
            **kwargs,
        )
        return

    def get_X_original(self):
        """Return the dataset with just the features to check."""
        return self.db[self.feat_to_check]

    def get_X_reduced(self):
        """Return the reduced dataset without the sample_col, target_col and the filtered features."""
        return self.db[self._selected_features]

    def get_db_filtered(self):
        """Return the reduced dataset without the filtered features."""
        return self.db.drop(columns=self._filtered_out_features)

    def make_funnel(self):
        """Return the funnel dataframe."""
        return self._funnel_df

    def make_report(self):
        """Return the feature selection report."""
        attr_list = [
            "_constant_features",
            "_missing_features",
            "_highly_concentrated_features",
            "_low_values_features",
            "_unstable_features",
            "_unexplanatory_features",
            "_correlated_features",
            "_collinear_features",
        ]
        report_list = []

        for attr in attr_list:
            report_list.append(
                pd.DataFrame({"feat_name": self.__dict__[attr], "result": "drop", "drop_reason": attr[1:-9]})
            )
        report_list.append(pd.DataFrame({"feat_name": self._selected_features, "result": "keep"}))

        return pd.concat(report_list).reset_index(drop=True)
